#include <sys/types.h>
#include <dirent.h>
#include <getopt.h>
#include <iostream>
#include <memory>
#include <fstream>
#include <cassert>
#include <set>
using namespace std;
#include "sitecing/sitecing_util.h"
#include "sitecing/util.h"
#include "sitecing/sitespace.h"
#include "sitecing/sitecing_interface_cgi.h"
#include "sitecing/cgi_component.h"
#include "sitecing/configuration.h"
#include "sitecing/magic.h"
#include "sitecing/sitecing_exception.h"
#include "sitecing/exception.h"
using namespace sitecing;

#include "config.h"
#define PHEADER PACKAGE "-build Version " VERSION
#define PCOPY   "Copyright (c) 2004 Klever Group"

static sitespace* site_space = NULL;
typedef pair<dev_t,ino_t> the_inode_t;
set<the_inode_t> built_inodes;

void build_component(const string& component) {
    assert(site_space);
    cerr << "Building " << component << endl;
    try {
	site_space->factory.make(site_space->config.root_so+component+".so");
    }catch(compile_error& ce) {
	ce.see(CODEPOINT);
	ifstream err((site_space->config.root_intermediate+ce.component_path+".stderr").c_str(),ios::in);
	if(err) {
	    cerr << err.rdbuf();
	}
	throw;
    }catch(preprocessor_error& pe) {
	pe.see(CODEPOINT);
	cerr << site_space->config.root_source << pe.component_name << ":" << pe.line_number << ":" << pe.what() << endl;
	throw;
    }
}

void build_imports(const string& component);

void build_with_imports(const string& component) {
    assert(site_space);
    struct stat st;
    string cp = site_space->config.root_source+component;
    if(!lstat(cp.c_str(),&st)) {
	if(built_inodes.find(the_inode_t(st.st_dev,st.st_ino))!=built_inodes.end())
	    return;
	built_inodes.insert(the_inode_t(st.st_dev,st.st_ino));
    }
    build_component(component);
    build_imports(component);
}

void build_imports(const string& component) {
    assert(site_space);
    ifstream ifs((site_space->config.root_intermediate+component+".imports").c_str(),ios::in);
    cerr << "Building components imported by " << component << endl;
    if(ifs) {
	string import;
	while(!ifs.eof()) {
	    ifs >> import;
	    if(!import.empty())
		build_with_imports(import);
	}
    }
}

void build_http_status_handlers(const string& target) {
    assert(site_space);
    set<string> stop_list;
    string t = "/";
    t += normalize_path(target,strip_leading_slash);
    for(;;) {
	if(t[t.length()-1]=='/') {
	    loaded_options* lo = site_space->config.lookup_loaded_options(t);
	    if( lo && (lo->flags&config_options::flag_http_status_handlers) ) {
		for(config_options::http_status_handlers_t::const_iterator i=lo->http_status_handlers.begin();i!=lo->http_status_handlers.end();++i) {
		    if(stop_list.find(i->first)==stop_list.end()) {
			build_with_imports(i->second);
			stop_list.insert(i->first);
		    }
		}
	    }
	}
	configuration::specs_t::iterator i=site_space->config.specs.find(t);
	if( i!=site_space->config.specs.end() && (i->second.flags&config_options::flag_http_status_handlers) ) {
	    for(config_options::http_status_handlers_t::const_iterator ii=i->second.http_status_handlers.begin();ii!=i->second.http_status_handlers.end();++i) {
		if(stop_list.find(ii->first)==stop_list.end()) {
		    build_with_imports(ii->second);
		    stop_list.insert(ii->first);
		}
	    }
	}
	if(t.empty())
	    return;
	string::size_type sl=t.rfind('/');
	if(sl==string::npos) {
	    t.erase();
	}else{
	    if(sl==(t.length()-1))
		t.erase(sl);
	    else
		t.erase(sl+1);
	}
    }
}

void build_target(const string& target) {
    assert(site_space);
    string action = target;
    config_options::action_handler_t *ah = site_space->config.lookup_action_handler(target);
    if(ah)
	action = ah->action;
    build_with_imports(action);
}

void build(const string& target) {
    assert(site_space);
    build_http_status_handlers(target);
    config_options *co_exception_handler = site_space->config.lookup_config(target,config_options::flag_exception_handler);
    if(co_exception_handler) {
	string handler = co_exception_handler->exception_handler;
	build_with_imports(handler);
    }
    string target_source = site_space->config.root_source+target;
    struct stat st;
    if(stat(target_source.c_str(),&st))
	throw konforka::exception(CODEPOINT,"failed to stat() target");
    if(S_ISREG(st.st_mode)) {
	build_target(target);
    }else if(S_ISDIR(st.st_mode)) {
	build_http_status_handlers(target);
	DIR *d=opendir(target_source.c_str());
        if(!d)
	    throw konforka::exception(CODEPOINT,"failed to opendir()");
	for(struct dirent *de=readdir(d);de;de=readdir(d)) {
	    if(!strcmp(de->d_name,"."))
		continue;
	    if(!strcmp(de->d_name,".."))
		continue;
	    string subtarget = normalize_path(target+"/"+de->d_name);
	    struct stat sts;
	    if(stat((site_space->config.root_source+subtarget).c_str(),&sts))
		throw konforka::exception(CODEPOINT,"failed to stat() subtarget");
	    if(S_ISDIR(sts.st_mode)) {
		build(subtarget);
	    }else{
		if(site_space->config.match_autobuild_files(target,de->d_name)){
		    build_target(subtarget);
		}
	    }
	}
	closedir(d);
    }
}

int main(int argc,char **argv) {
    try {
	string config_file = "sitecing.conf";
	while(true) {
	    static struct option opts[] = {
		{ "help", no_argument, 0, 'h' },
		{ "usage", no_argument, 0, 'h' },
		{ "version", no_argument, 0, 'V' },
		{ "license", no_argument, 0, 'L' },
		{ "config", required_argument, 0, 'f' },
		{ NULL, 0, 0, 0 }
	    };
	    int c = getopt_long(argc,argv,"f:hVL",opts,NULL);
	    if(c==-1)
		break;
	    switch(c) {
		case 'h':
		    cerr << PHEADER << endl
			<< PCOPY << endl << endl
			<< " -h, --help" << endl
			<< " --usage          display this text" << endl
			<< " -V, --version    display version number" << endl
			<< " -L, --license    show license" << endl
			<< " -f filename, --config=filename" << endl
			<< "                  specify configuration file to use" << endl;
		    exit(0);
		    break;
		case 'V':
		    cerr << VERSION << endl;
		    exit(0);
		    break;
		case 'L':
		    extern const char *COPYING;
		    cerr << COPYING << endl;
		    exit(0);
		    break;
		case 'f':
		    config_file = optarg;
		    break;
		default:
		    cerr << "Huh??" << endl;
		    break;
	    }
	}
	configuration config(config_file,true);
	if(!(config.flags&configuration::flag_root_source))
	    throw konforka::exception(CODEPOINT,"Unspecified root for sources");
	if(!(config.flags&configuration::flag_root_intermediate))
	    throw konforka::exception(CODEPOINT,"Unspecified root for intermediate files");
	if(!(config.flags&configuration::flag_root_so))
	    throw konforka::exception(CODEPOINT,"Unspecified root for shared objects");
	sitespace ss(config);
	site_space = &ss;
	if(optind<argc) {
	    for(int narg=optind;narg<argc;narg++) {
		const char *arg = argv[narg];
		build(arg);
	    }
	}else
	    build("/");
	return 0;
    }catch(exception& e) {
	cerr << "Oops: " << e.what() << endl;
	return 1;
    }
}
